# coding: utf-8

# docs
from __future__ import annotations

import asyncio
import json
import re
from collections import defaultdict
from typing import (
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import numpy as np
from requests_toolbelt import MultipartDecoder, MultipartEncoder
from tqdm import tqdm

from supervisely._utils import batched, logger, run_coroutine
from supervisely.annotation.label import LabelingStatus
from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.geometry.rectangle import Rectangle
from supervisely.video_annotation.key_id_map import KeyIdMap


class FigureInfo(NamedTuple):
    """
    Represents detailed information about a figure in a scene within the labeling tool.
    It is designed to handle multimodal data.

    Attributes:
        id (int): Unique identifier for the figure.
        class_id (int): Identifier for the class of the figure.
        updated_at (str): Timestamp of the last update.
        created_at (str): Timestamp of creation.
        entity_id (int): Identifier for the entity. Possible entities: image, video, volume, pointcloud, etc.
        object_id (int): Identifier for the object (applicable to videos, volumes, pointclouds).
        project_id (int): Identifier for the project.
        dataset_id (int): Identifier for the dataset.
        frame_index (int): Index of the frame (applicable to videos, volumes (as a slice_index), pointclouds).
        geometry_type (str): Type of geometry.
        geometry (dict): Geometry data.
        geometry_meta (dict): Metadata for the geometry.
        tags (list): List of tags associated with the figure.
        meta (dict): Additional metadata.
        area (str): Area information.
        priority (int): Position of the figure relative to other overlapping or underlying figures.
    """

    id: int
    class_id: int
    updated_at: str
    created_at: str
    entity_id: int
    object_id: int
    project_id: int
    dataset_id: int
    frame_index: int
    geometry_type: str
    geometry: dict
    geometry_meta: dict
    tags: list
    meta: dict
    area: str
    priority: Optional[int] = None
    custom_data: Optional[dict] = None

    @property
    def bbox(self) -> Optional[Rectangle]:
        """
        Get Figure's bounding box.

        :returns: Rectangle in supervisely format.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`
        """
        if self.geometry_meta is not None:
            return Rectangle(*self.geometry_meta["bbox"], sly_id=self.id)

    def to_json(self):
        return FigureApi.convert_info_to_json(self)


class FigureApi(RemoveableBulkModuleApi):
    """
    Figure object for :class:`~supervisely.video_annotation.video_annotation.VideoAnnotation`.
    """

    def _remove_batch_api_method_name(self):
        """_remove_batch_api_method_name"""
        return "figures.bulk.remove"

    def _remove_batch_field_name(self):
        """_remove_batch_field_name"""
        return ApiField.FIGURE_IDS

    @staticmethod
    def info_sequence():
        """
        NamedTuple FigureInfo information about Figure.

        :Example:

         .. code-block:: python

            FigureInfo(id=588801373,
                       updated_at='2020-12-22T06:37:13.183Z',
                       created_at='2020-12-22T06:37:13.183Z',
                       entity_id=186648101,
                       object_id=112482,
                       project_id=110366,
                       dataset_id=419886,
                       frame_index=0,
                       geometry_type='bitmap',
                       geometry={'bitmap': {'data': 'eJwdlns8...Cgj4=', 'origin': [335, 205]}})
        """
        return [
            ApiField.ID,
            ApiField.CLASS_ID,
            ApiField.UPDATED_AT,
            ApiField.CREATED_AT,
            ApiField.ENTITY_ID,
            ApiField.OBJECT_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.FRAME_INDEX,
            ApiField.GEOMETRY_TYPE,
            ApiField.GEOMETRY,
            ApiField.GEOMETRY_META,
            ApiField.TAGS,
            ApiField.META,
            ApiField.AREA,
            ApiField.PRIORITY,
            ApiField.CUSTOM_DATA,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :returns: NamedTuple name.
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                tuple_name = api.video.figure.info_tuple_name()
                print(tuple_name) # FigureInfo
        """

        return "FigureInfo"

    def get_info_by_id(self, id: int) -> FigureInfo:
        """
        Get Figure information by ID.

        :param id: Figure ID in Supervisely.
        :type id: int
        :returns: Information about Figure.
        :rtype: :class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                figure_id = 588801373
                figure_info = api.video.figure.get_info_by_id(figure_id)
                print(figure_info)
                # Output: [
                #     588801373,
                #     "2020-12-22T06:37:13.183Z",
                #     "2020-12-22T06:37:13.183Z",
                #     186648101,
                #     112482,
                #     110366,
                #     419886,
                #     0,
                #     "bitmap",
                #     {
                #         "bitmap": {
                #             "data": "eJw...Cgj4=",
                #             "origin": [
                #                 335,
                #                 205
                #             ]
                #         }
                #     }
                # ]
        """
        fields = [
            "id",
            "createdAt",
            "updatedAt",
            "imageId",
            "objectId",
            "classId",
            "projectId",
            "datasetId",
            "geometry",
            "geometryType",
            "geometryMeta",
            "tags",
            "meta",
            "area",
            "priority",
            "nnCreated",
            "nnUpdated",
        ]
        return self._get_info_by_id(id, "figures.info", {ApiField.FIELDS: fields})

    def create(
        self,
        entity_id: int,
        object_id: int,
        meta: Dict,
        geometry_json: Dict,
        geometry_type: str,
        track_id: Optional[int] = None,
        custom_data: Optional[dict] = None,
        status: Optional[LabelingStatus] = None,
    ) -> int:
        """"""
        input_figure = {
            ApiField.META: meta,
            ApiField.OBJECT_ID: object_id,
            ApiField.GEOMETRY_TYPE: geometry_type,
            ApiField.GEOMETRY: geometry_json,
        }

        if status is None:
            status = LabelingStatus.MANUAL

        nn_created, nn_updated = LabelingStatus.to_flags(status)
        input_figure[ApiField.NN_CREATED] = nn_created
        input_figure[ApiField.NN_UPDATED] = nn_updated

        if track_id is not None:
            input_figure[ApiField.TRACK_ID] = track_id

        if custom_data is not None:
            input_figure[ApiField.CUSTOM_DATA] = custom_data

        body = {ApiField.ENTITY_ID: entity_id, ApiField.FIGURES: [input_figure]}

        response = self._api.post("figures.bulk.add", body)
        return response.json()[0][ApiField.ID]

    def get_by_ids(self, dataset_id: int, ids: List[int]) -> List[FigureInfo]:
        """
        Get Figures information by IDs from given dataset ID.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param ids: List of Figures IDs.
        :type ids: List[int]
        :returns: List of information about Figures.
        :rtype: List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 466642
                figures_ids = [642155547, 642155548, 642155549]
                figures_infos = api.video.figure.get_by_ids(dataset_id, figures_ids)
                print(figures_infos)
                # Output: [
                #     [
                #         642155547,
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         198703211,
                #         152118,
                #         124976,
                #         466642,
                #         0,
                #         "rectangle",
                #         {
                #             "points": {
                #                 "exterior": [
                #                     [
                #                         2240,
                #                         1041
                #                     ],
                #                     [
                #                         2463,
                #                         1187
                #                     ]
                #                 ],
                #                 "interior": []
                #             }
                #         }
                #     ],
                #     [
                #         642155548,
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         198703211,
                #         152118,
                #         124976,
                #         466642,
                #         1,
                #         "rectangle",
                #         {
                #             "points": {
                #                 "exterior": [
                #                     [
                #                         2248,
                #                         1048
                #                     ],
                #                     [
                #                         2455,
                #                         1176
                #                     ]
                #                 ],
                #                 "interior": []
                #             }
                #         }
                #     ],
                #     [
                #         642155549,
                #         "2021-03-23T13:25:34.705Z",
                #         "2021-03-23T13:25:34.705Z",
                #         198703211,
                #         152118,
                #         124976,
                #         466642,
                #         2,
                #         "rectangle",
                #         {
                #             "points": {
                #                 "exterior": [
                #                     [
                #                         2237,
                #                         1046
                #                     ],
                #                     [
                #                         2464,
                #                         1179
                #                     ]
                #                 ],
                #                 "interior": []
                #             }
                #         }
                #     ]
                # ]
        """
        filters = [{"field": "id", "operator": "in", "value": ids}]
        fields = [
            ApiField.ID,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.IMAGE_ID,
            ApiField.OBJECT_ID,
            ApiField.CLASS_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.GEOMETRY,
            ApiField.GEOMETRY_TYPE,
            ApiField.GEOMETRY_META,
            ApiField.TAGS,
            ApiField.META,
            ApiField.AREA,
            ApiField.PRIORITY,
            ApiField.CUSTOM_DATA,
            ApiField.NN_CREATED,
            ApiField.NN_UPDATED,
        ]
        figures_infos = self.get_list_all_pages(
            "figures.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters,
                ApiField.FIELDS: fields,
            },
        )

        if len(ids) != len(figures_infos):
            ids_downloaded = [info.id for info in figures_infos]
            raise RuntimeError(
                "Ids don't exist on server: {}".format(set(ids_downloaded) - set(ids))
            )

        id_to_item = {info.id: info for info in figures_infos}

        figures = []
        for input_id in ids:
            figures.append(id_to_item[input_id])

        return figures

    def _append_bulk(
        self,
        entity_id,
        figures_json,
        figures_keys,
        key_id_map: KeyIdMap,
        field_name=ApiField.ENTITY_ID,
    ):
        """"""
        if len(figures_json) == 0:
            return
        for batch_keys, batch_jsons in zip(
            batched(figures_keys, batch_size=100), batched(figures_json, batch_size=100)
        ):
            resp = self._api.post(
                "figures.bulk.add",
                {field_name: entity_id, ApiField.FIGURES: batch_jsons},
            )
            for key, resp_obj in zip(batch_keys, resp.json()):
                figure_id = resp_obj[ApiField.ID]
                key_id_map.add_figure(key, figure_id)

    def create_bulk(
        self,
        figures_json: List[dict],
        entity_id: int = None,
        dataset_id: int = None,
        batch_size=200,
    ) -> List[int]:
        """
        Create figures in Supervisely in bulk.
        To optimize creation of a large number of figures use dataset ID instead of entity ID.
        In this case figure jsons list can contain figures from different entities for the same dataset.
        Every figure json must contain corresponding entity ID.

        *NOTE*: Geometries for AlphaMask must be uploaded separately via `upload_geometries_batch` method.

        :param figures_json: List of figures in Supervisely JSON format.
        :type figures_json: List[dict]
        :param entity_id: Entity ID.
        :type entity_id: int
        :param dataset_id: Dataset ID.
        :type dataset_id: int
        :returns: List of figure IDs.
        """
        figure_ids = []
        if len(figures_json) == 0:
            return figure_ids

        if entity_id is None and dataset_id is None:
            raise ValueError("Either entity_id or dataset_id must be provided")
        if dataset_id is not None:
            body = {ApiField.DATASET_ID: dataset_id}
        else:
            body = {ApiField.ENTITY_ID: entity_id}

        for batch_jsons in batched(figures_json, batch_size):
            body[ApiField.FIGURES] = batch_jsons
            response = self._api.post("figures.bulk.add", body)
            for resp_obj in response.json():
                figure_ids.append(resp_obj[ApiField.ID])
        return figure_ids

    def download(
        self, dataset_id: int, image_ids: List[int] = None, skip_geometry: bool = False
    ) -> Dict[int, List[FigureInfo]]:
        """
        Method returns a dictionary with pairs of image ID and list of FigureInfo for the given dataset ID. Can be filtered by image IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: Specify the list of image IDs within the given dataset ID. If image_ids is None, the method returns all possible pairs of images with figures. Note: Consider using `sly.batched()` to ensure that no figures are lost in the response.
        :type image_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool

        :returns: A dictionary where keys are image IDs and values are lists of figures.
        :rtype: Dict[int, List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]]
        """
        fields = [
            ApiField.ID,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.IMAGE_ID,
            ApiField.OBJECT_ID,
            ApiField.CLASS_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.GEOMETRY,
            ApiField.GEOMETRY_TYPE,
            ApiField.GEOMETRY_META,
            ApiField.TAGS,
            ApiField.META,
            ApiField.AREA,
            ApiField.PRIORITY,
            ApiField.CUSTOM_DATA,
            ApiField.NN_CREATED,
            ApiField.NN_UPDATED,
        ]
        if skip_geometry is True:
            fields = [x for x in fields if x != ApiField.GEOMETRY]

        if image_ids is None:
            filters = []
        else:
            filters = [
                {
                    ApiField.FIELD: ApiField.ENTITY_ID,
                    ApiField.OPERATOR: "in",
                    ApiField.VALUE: image_ids,
                }
            ]
        data = {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FIELDS: fields,
            ApiField.FILTER: filters,
        }
        resp = self._api.post("figures.list", data)
        infos = resp.json()
        images_figures = defaultdict(list)

        total_pages = infos["pagesCount"]
        for page in range(1, total_pages + 1):
            if page > 1:
                data.update({ApiField.PAGE: page})
                resp = self._api.post("figures.list", data)
                infos = resp.json()
            for info in infos["entities"]:
                figure_info = self._convert_json_info(info, True)
                images_figures[figure_info.entity_id].append(figure_info)

        return dict(images_figures)

    def _convert_json_info(self, info: dict, skip_missing=False):
        res = super()._convert_json_info(info, skip_missing=True)
        return FigureInfo(**res._asdict())

    def _download_geometries_generator(
        self, ids: List[int]
    ) -> Generator[Tuple[int, MultipartDecoder.Part], None, None]:
        """
        Private method. Download figures geometries with given IDs from storage.
        """

        for batch_ids in batched(ids):
            response = self._api.post("figures.bulk.download.geometry", {ApiField.IDS: batch_ids})
            decoder = MultipartDecoder.from_response(response)
            for part in decoder.parts:
                content_utf8 = part.headers[b"Content-Disposition"].decode("utf-8")
                # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                # The regex has 2 capture group: one for the prefix and one for the actual name value.
                figure_id = int(re.findall(r'(^|[\s;])name="(\d*)"', content_utf8)[0][1])
                yield figure_id, part

    def download_geometry(self, figure_id: int) -> dict:
        """
        Download figure geometry with given ID from storage.

        :param figure_id: Figure ID in Supervisely.
        :type figure_id: int
        :returns: Figure geometry in Supervisely JSON format.
        :rtype: dict
        """
        return self.download_geometries_batch([figure_id])

    def download_geometries_batch(
        self,
        ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> List[dict]:
        """
        Download figure geometries with given IDs from storage.

        :param ids: List of figure IDs in Supervisely.
        :type ids: List[int]
        :param progress_cb: Progress bar to show the download progress. Shows the number of bytes downloaded.
        :type progress_cb: Union[tqdm, Callable], optional
        :returns: List of figure geometries in Supervisely JSON format.
        :rtype: List[dict]
        """
        geometries = {}
        for idx, part in self._download_geometries_generator(ids):
            try:
                if progress_cb is not None:
                    progress_cb(len(part.content))
                geometry_json = json.loads(part.content)
                geometries[idx] = geometry_json
            except Exception as e:
                raise RuntimeError(f"Failed to decode geometry for figure ID {idx}") from e

        if len(geometries) != len(ids):
            raise RuntimeError("Not all geometries were downloaded")
        ordered_results = [geometries[i] for i in ids]
        return ordered_results

    def upload_geometry(self, figure_id: int, geometry: dict):
        """
        Upload figure geometry with given figure ID to storage.

        :param figure_id: Figure ID in Supervisely.
        :type figure_id: int
        :param geometry: Figure geometry in Supervisely JSON format.
        :type geometry: dict
        :returns: None
        :rtype: None
        """
        self.upload_geometries_batch([figure_id], [geometry])

    def upload_geometries_batch(self, figure_ids: List[int], geometries: List[dict]):
        """
        Upload figure geometries with given figure IDs to storage.

        :param figure_ids: List of figure IDs in Supervisely.
        :type figure_ids: List[int]
        :param geometries: List of figure geometries in Supervisely JSON format.
        :type geometries: List[dict]
        :returns: None
        :rtype: None
        """
        geometries = [json.dumps(geometry).encode("utf-8") for geometry in geometries]

        for batch_ids, batch_geometries in zip(
            batched(figure_ids, batch_size=100), batched(geometries, batch_size=100)
        ):
            fields = []
            for figure_id, geometry in zip(batch_ids, batch_geometries):
                fields.append((ApiField.FIGURE_ID, str(figure_id)))
                fields.append(
                    (
                        ApiField.GEOMETRY,
                        (str(figure_id), geometry, "application/octet-stream"),
                    )
                )
            encoder = MultipartEncoder(fields=fields)
            self._api.post("figures.bulk.upload.geometry", encoder)

    async def _download_geometries_generator_async(
        self, ids: List[int], semaphore: Optional[asyncio.Semaphore] = None
    ) -> AsyncGenerator[Tuple[int, bytes], None, None]:
        """
        Private method. Download figures geometries with given IDs from storage asynchronously.

        :param ids: List of figure IDs in Supervisely.
        :type ids: List[int]
        :returns: Async generator with pairs of figure ID and figure geometry.
        :rtype: AsyncGenerator[Tuple[int, bytes], None, None]
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        for batch_ids in batched(ids):
            async with semaphore:
                response = await self._api.post_async(
                    "figures.bulk.download.geometry", {ApiField.IDS: batch_ids}
                )
                decoder = MultipartDecoder.from_response(response)
                for part in decoder.parts:
                    content_utf8 = part.headers[b"Content-Disposition"].decode("utf-8")
                    # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                    # The regex has 2 capture group: one for the prefix and one for the actual name value.
                    figure_id = int(re.findall(r'(^|[\s;])name="(\d*)"', content_utf8)[0][1])
                    yield figure_id, part.content

    async def download_geometries_batch_async(
        self,
        ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        semaphore: Optional[asyncio.Semaphore] = None,
    ) -> List[dict]:
        """
        Download figure geometries with given IDs from storage asynchronously.

        :param ids: List of figure IDs in Supervisely.
        :type ids: List[int]
        :param progress_cb: Progress bar to show the download progress. Shows the number of bytes downloaded.
        :type progress_cb: Union[tqdm, Callable], optional
        :param semaphore: Semaphore to limit the number of concurrent downloads.
        :type semaphore: Optional[asyncio.Semaphore], optional
        :returns: List of figure geometries in Supervisely JSON format.
        :rtype: List[dict]

        :Usage Example:

            .. code-block:: python

                import os
                import asyncio
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                figure_ids = [642155547, 642155548, 642155549]
                loop = sly.utils.get_or_create_event_loop()
                geometries = loop.run_until_complete(
                    api.figure.download_geometries_batch_async(
                        figure_ids,
                        progress_cb=tqdm(total=len(figure_ids), desc="Downloading geometries"),
                        semaphore=asyncio.Semaphore(15),
                    )
                )
        """
        geometries = {}
        async for idx, part in self._download_geometries_generator_async(ids, semaphore):
            if progress_cb is not None:
                progress_cb(len(part))
            geometry_json = json.loads(part)
            geometries[idx] = geometry_json

        if len(geometries) != len(ids):
            raise RuntimeError("Not all geometries were downloaded")
        ordered_results = [geometries[i] for i in ids]
        return ordered_results

    async def upload_geometries_batch_async(
        self,
        figure_ids: List[int],
        geometries: List[dict],
        semaphore: Optional[asyncio.Semaphore] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """
        Upload figure geometries with given figure IDs to storage asynchronously in batches.

        :param figure_ids: List of figure IDs in Supervisely.
        :type figure_ids: List[int]
        :param geometries: List of figure geometries in Supervisely JSON format.
        :type geometries: List[dict]
        :param semaphore: Semaphore to limit the number of concurrent uploads.
        :type semaphore: Optional[asyncio.Semaphore], optional
        :param progress_cb: Progress bar to show the upload progress. Shows the number of geometries uploaded.
        :type progress_cb: Union[tqdm, Callable], optional
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                import asyncio
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                figure_ids = [642155547, 642155548, 642155549]
                geometries = [{...}, {...}, {...}]  # Your geometry data

                upload_coroutine = api.figure.upload_geometries_batch_async(
                    figure_ids,
                    geometries,
                    semaphore=asyncio.Semaphore(10),
                )
                sly.run_coroutine(upload_coroutine)
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        encoded_geometries = [json.dumps(geometry).encode("utf-8") for geometry in geometries]

        batch_size = 100
        tasks = []

        for batch_ids, batch_geometries in zip(
            batched(figure_ids, batch_size),
            batched(encoded_geometries, batch_size),
        ):
            fields = []
            for figure_id, geometry in zip(batch_ids, batch_geometries):
                fields.append((ApiField.FIGURE_ID, str(figure_id)))
                fields.append(
                    (
                        ApiField.GEOMETRY,
                        (str(figure_id), geometry, "application/octet-stream"),
                    )
                )

            async def upload_batch(fields, progress_cb, num):
                async with semaphore:
                    encoder = MultipartEncoder(fields=fields)
                    headers = {"Content-Type": encoder.content_type}
                    async for _, _ in self._api.stream_async(
                        "figures.bulk.upload.geometry",
                        "POST",
                        data=encoder,
                        content=encoder.to_string(),
                        headers=headers,
                    ):
                        pass
                if progress_cb is not None:
                    progress_cb.update(num)

            tasks.append(upload_batch(fields, progress_cb, len(batch_ids)))

        if tasks:
            await asyncio.gather(*tasks)

    async def download_async(
        self,
        dataset_id: int,
        image_ids: Optional[List[int]] = None,
        skip_geometry: bool = False,
        semaphore: Optional[asyncio.Semaphore] = None,
        log_progress: bool = True,
        batch_size: int = 300,
    ) -> Dict[int, List[FigureInfo]]:
        """
        Asynchronously download figures for the given dataset ID. Can be filtered by image IDs.
        This method is significantly faster than the synchronous version for large datasets.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: Specify the list of image IDs within the given dataset ID. If image_ids is None, the method returns all possible pairs of images with figures.
        :type image_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool
        :param semaphore: Semaphore to limit the number of concurrent downloads.
        :type semaphore: Optional[asyncio.Semaphore], optional
        :param log_progress: If True, log the progress of the download.
        :type log_progress: bool, optional
        :param batch_size: Size of the batch for downloading figures per 1 request. Default is 300.
                        Used for batching image_ids when filtering by specific images.
                        Adjust this value for optimal performance, value cannot exceed 500.
        :type batch_size: int, optional
        :returns: A dictionary where keys are image IDs and values are lists of figures.
        :rtype: Dict[int, List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 12345
                download_coroutine = api.image.figure.download_async(dataset_id)
                figures = sly.run_coroutine(download_coroutine)
        """
        fields = [
            ApiField.ID,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.IMAGE_ID,
            ApiField.OBJECT_ID,
            ApiField.CLASS_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.GEOMETRY,
            ApiField.GEOMETRY_TYPE,
            ApiField.GEOMETRY_META,
            ApiField.TAGS,
            ApiField.META,
            ApiField.AREA,
            ApiField.PRIORITY,
            ApiField.CUSTOM_DATA,
            ApiField.NN_CREATED,
            ApiField.NN_UPDATED,
        ]
        if skip_geometry is True:
            fields = [x for x in fields if x != ApiField.GEOMETRY]

        # Base data setup
        base_data = {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FIELDS: fields,
        }

        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        async def _get_page_figures(page_data, semaphore, progress_cb: tqdm = None):
            """Helper function to get figures from a single page"""
            async with semaphore:
                response = await self._api.post_async("figures.list", page_data)
                response_json = response.json()

                page_figures = []
                for info in response_json["entities"]:
                    figure_info = self._convert_json_info(info, True)
                    page_figures.append(figure_info)
                if progress_cb is not None:
                    progress_cb.update(len(response_json["entities"]))
                return page_figures

        async def _get_all_pages(ids_filter, progress_cb: tqdm = None):
            """Internal function to process all pages for given filter"""
            data = base_data.copy()
            data[ApiField.FILTER] = ids_filter

            # Get first page to determine pagination
            data[ApiField.PAGE] = 1
            async with semaphore:
                response = await self._api.post_async("figures.list", data)
                response_json = response.json()

            pages_count = response_json["pagesCount"]
            all_figures = []

            # Process first page
            for info in response_json["entities"]:
                figure_info = self._convert_json_info(info, True)
                all_figures.append(figure_info)
            if progress_cb is not None:
                progress_cb.update(len(response_json["entities"]))

            # Process remaining pages in parallel if needed
            if pages_count > 1:
                tasks = []
                for page in range(2, pages_count + 1):
                    page_data = data.copy()
                    page_data[ApiField.PAGE] = page
                    tasks.append(
                        asyncio.create_task(
                            _get_page_figures(page_data, semaphore, progress_cb=progress_cb)
                        )
                    )

                if tasks:
                    page_results = await asyncio.gather(*tasks)
                    for page_figures in page_results:
                        all_figures.extend(page_figures)

            return all_figures

        if log_progress:
            progress_cb = tqdm(desc="Downloading figures", unit="figure", total=0)
        else:
            progress_cb = None

        # Strategy: batch processing based on image_ids
        tasks = []

        if image_ids is None:
            # Single task for all figures in dataset
            filters = []
            tasks.append(_get_all_pages(filters, progress_cb=progress_cb))
        else:
            # Batch image_ids and create tasks for each batch
            for batch_ids in batched(image_ids, batch_size):
                filters = [
                    {
                        ApiField.FIELD: ApiField.ENTITY_ID,
                        ApiField.OPERATOR: "in",
                        ApiField.VALUE: list(batch_ids),
                    }
                ]
                tasks.append(_get_all_pages(filters, progress_cb=progress_cb))
                # Small delay between batches to reduce server load
                await asyncio.sleep(0.02)

        # Execute all tasks in parallel and collect results
        all_results = await asyncio.gather(*tasks)

        # Combine results from all batches
        images_figures = defaultdict(list)

        for batch_figures in all_results:
            for figure in batch_figures:
                images_figures[figure.entity_id].append(figure)

        return dict(images_figures)

    def download_fast(
        self,
        dataset_id: int,
        image_ids: Optional[List[int]] = None,
        skip_geometry: bool = False,
        semaphore: Optional[asyncio.Semaphore] = None,
        log_progress: bool = True,
        batch_size: int = 300,
    ) -> Dict[int, List[FigureInfo]]:
        """
        Download figures for the given dataset ID. Can be filtered by image IDs.
        This method is significantly faster than the synchronous version for large datasets and
        is designed to be used in an asynchronous context.
        Will fallback to the synchronous version if async fails.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: Specify the list of image IDs within the given dataset ID. If image_ids is None, the method returns all possible pairs of images with figures.
        :type image_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool
        :param semaphore: Semaphore to limit the number of concurrent downloads.
        :type semaphore: Optional[asyncio.Semaphore], optional
        :param log_progress: If True, log the progress of the download.
        :type log_progress: bool, optional
        :param batch_size: Size of the batch for downloading figures per 1 request. Default is 300.
                        Used for batching image_ids when filtering by specific images.
                        Adjust this value for optimal performance, value cannot exceed 500.
        :type batch_size: int, optional

        :returns: A dictionary where keys are image IDs and values are lists of figures.
        :rtype: Dict[int, List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 12345
                figures = api.image.figure.download_fast(dataset_id)
        """
        try:
            return run_coroutine(
                self.download_async(
                    dataset_id=dataset_id,
                    image_ids=image_ids,
                    skip_geometry=skip_geometry,
                    semaphore=semaphore,
                    log_progress=log_progress,
                    batch_size=batch_size,
                )
            )
        except Exception:
            # Fallback to the synchronous version if async fails
            logger.debug("Async download of figures is failed, falling back to sync download.")
            if log_progress:
                logger.debug("Downloading figures without progress bar.")
            return self.download(
                dataset_id=dataset_id,
                image_ids=image_ids,
                skip_geometry=skip_geometry,
            )

    def restore_batch(self, ids: List[int], progress_cb: Optional[Callable] = None, batch_size: int = 50):
        """
        Restore archived figures in batches from the Supervisely server.

        :param ids: IDs of figures in Supervisely.
        :type ids: List[int]
        :param progress_cb: Optional callback to track restore progress. Receives number of restored figures in the current batch.
        :type progress_cb: Optional[Callable]
        :param batch_size: Number of figure IDs to send in a single request.
        :type batch_size: int
        """
        for ids_batch in batched(ids, batch_size=batch_size):
            self._api.post(
                "figures.bulk.restore",
                {ApiField.FIGURE_IDS: ids_batch},
            )
            if progress_cb is not None:
                progress_cb(len(ids_batch))

    def restore(self, id: int):
        """
        Restore a single archived figure with the specified ID from the Supervisely server.

        :param id: Figure ID in Supervisely.
        :type id: int
        """
        self.restore_batch([id])



from supervisely.nn.inference.inference import Inference, Timer
from supervisely.nn.inference.gui.gui import InferenceGUI, BaseInferenceGUI
from supervisely.nn.inference.instance_segmentation.instance_segmentation import (
    InstanceSegmentation,
)
from supervisely.nn.inference.object_detection.object_detection import ObjectDetection
from supervisely.nn.inference.semantic_segmentation.semantic_segmentation import (
    SemanticSegmentation,
)
from supervisely.nn.inference.pose_estimation.pose_estimation import PoseEstimation
from supervisely.nn.inference.salient_object_segmentation.salient_object_segmentation import (
    SalientObjectSegmentation,
)
from supervisely.nn.inference.prompt_based_object_detection.propmt_based_object_detection import (
    PromptBasedObjectDetection,
)
from supervisely.nn.inference.interactive_segmentation.interactive_segmentation import (
    InteractiveSegmentation,
)
from supervisely.nn.inference.promptable_segmentation.promptable_segmentation import (
    PromptableSegmentation,
)
from supervisely.nn.inference.session import Session, SessionJSON
from supervisely.nn.inference.tracking.point_tracking import PointTracking
from supervisely.nn.inference.tracking.bbox_tracking import BBoxTracking
from supervisely.nn.inference.tracking.mask_tracking import MaskTracking
from supervisely.nn.inference.tracking.object_tracking_3d import ObjectTracking3D
from supervisely.nn.inference.object_detection_3d import ObjectDetection3D
from supervisely.nn.utils import CheckpointInfo, DeployInfo, RuntimeType, ModelSource, ModelPrecision
from supervisely.nn.task_type import TaskType



# coding: utf-8
"""Create, monitor, and manage Supervisely tasks."""

import json
import os
import time
from collections import OrderedDict, defaultdict
from pathlib import Path

# docs
from typing import Any, Callable, Dict, List, Literal, NamedTuple, Optional, Union

import requests
from pydantic import BaseModel, Field
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from supervisely import logger
from supervisely._utils import batched, take_with_default
from supervisely.api.module_api import (
    ApiField,
    ModuleApiBase,
    ModuleWithStatus,
    WaitingTimeExceeded,
)
from supervisely.collection.str_enum import StrEnum
from supervisely.io.env import app_categories
from supervisely.io.fs import (
    ensure_base_path,
    get_file_hash,
    get_file_name,
    get_file_name_with_ext,
)


class KubernetesSettings(BaseModel):
    """Application resource limits and requests (CPUs, memory, GPU, storage) for tasks."""

    use_health_check: Optional[bool] = Field(None, alias="useHealthCheck")
    request_cpus: Optional[int] = Field(None, alias="requestCpus")
    limit_cpus: Optional[int] = Field(None, alias="limitCpus")
    limit_memory_gb: Optional[int] = Field(None, alias="limitMemoryGb")
    limit_shm_gb: Optional[int] = Field(None, alias="limitShmGb")
    limit_storage_gb: Optional[int] = Field(None, alias="limitStorageGb")
    limit_gpus: Optional[int] = Field(None, alias="limitGpus")
    limit_gpu_memory_mb: Optional[int] = Field(None, alias="limitGpuMemoryMb")
    limit_gpu_cores_perc: Optional[int] = Field(None, alias="limitGpuCoresPerc")

    model_config = {"populate_by_name": True}

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dict with only non-None values using aliases."""
        return self.model_dump(exclude_none=True, by_alias=True)


class TaskFinishedWithError(Exception):
    """Raised when a task finishes in an error state while waiting/polling."""

    pass


class TaskApi(ModuleApiBase, ModuleWithStatus):
    """API for working with tasks."""

    class RestartPolicy(StrEnum):
        """Task restart policy used for app deployments."""

        NEVER = "never"
        """"""
        ON_ERROR = "on_error"
        """"""

    class PluginTaskType(StrEnum):
        """Task type for plugin-based tasks (train/inference/smarttool/etc.)."""

        TRAIN = "train"
        """"""
        INFERENCE = "inference"
        """"""
        INFERENCE_RPC = "inference_rpc"
        """"""
        SMART_TOOL = "smarttool"
        """"""
        CUSTOM = "custom"
        """"""

    class Status(StrEnum):
        """Task lifecycle status values returned by the API."""

        QUEUED = "queued"
        """Application is queued for execution"""
        CONSUMED = "consumed"
        """Application is consumed by an agent"""
        STARTED = "started"
        """Application has been started"""
        DEPLOYED = "deployed"
        """Only for Plugins"""
        ERROR = "error"
        """Application has finished with an error"""
        FINISHED = "finished"
        """Application has finished successfully"""
        TERMINATING = "terminating"
        """Application is being terminated"""
        STOPPED = "stopped"
        """Application has been stopped"""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                task_info = api.task.get_info_by_id(task_id)
        """
        ModuleApiBase.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(
        self, workspace_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of Application Tasks in the given Workspace.

        :param workspace_id: Workspace ID.
        :type workspace_id: int
        :param filters: List of params to sort output Projects.
        :type filters: List[dict], optional
        :returns: List of Tasks with information for the given Workspace.
        :rtype: :class:`List[NamedTuple]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                workspace_id = 23821
                task_infos = api.task.get_list(workspace_id)
                task_infos_filter = api.task.get_list(23821, filters=[{'field': 'id', 'operator': '=', 'value': 121230}])
                print(task_infos_filter)
                # Output: [
                #     {
                #         "id": 121230,
                #         "type": "clone",
                #         "status": "finished",
                #         "startedAt": "2019-12-19T12:13:09.702Z",
                #         "finishedAt": "2019-12-19T12:13:09.701Z",
                #         "meta": {
                #             "input": {
                #                 "model": {
                #                     "id": 1849
                #                 },
                #                 "isExternal": true,
                #                 "pluginVersionId": 84479
                #             },
                #             "output": {
                #                 "model": {
                #                     "id": 12380
                #                 },
                #                 "pluginVersionId": 84479
                #             }
                #         },
                #         "description": ""
                #     }
                # ]
        """
        return self.get_list_all_pages(
            "tasks.list",
            {ApiField.WORKSPACE_ID: workspace_id, ApiField.FILTER: filters or []},
        )

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Application Task information by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :returns: Information about Task.
        :rtype: :class:`NamedTuple`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                task_id = 121230
                task_info = api.task.get_info_by_id(task_id)
                print(task_info)
                # Output: {
                #     "id": 121230,
                #     "workspaceId": 23821,
                #     "description": "",
                #     "type": "clone",
                #     "status": "finished",
                #     "startedAt": "2019-12-19T12:13:09.702Z",
                #     "finishedAt": "2019-12-19T12:13:09.701Z",
                #     "userId": 16154,
                #     "meta": {
                #         "app": {
                #             "id": 10370,
                #             "name": "Auto Import",
                #             "version": "test-branch",
                #             "isBranch": true,
                #         },
                #         "input": {
                #             "model": {
                #                 "id": 1849
                #             },
                #             "isExternal": true,
                #             "pluginVersionId": 84479
                #         },
                #         "output": {
                #             "model": {
                #                 "id": 12380
                #             },
                #             "pluginVersionId": 84479
                #         }
                #     },
                #     "settings": {},
                #     "agentName": null,
                #     "userLogin": "alexxx",
                #     "teamId": 16087,
                #     "agentId": null
                # }
        """
        return self._get_info_by_id(id, "tasks.info")

    def get_status(self, task_id: int) -> Status:
        """
        Check status of Application Task by ID.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :returns: Status object
        :rtype: :class:`Status`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                task_id = 121230
                task_status = api.task.get_status(task_id)
                print(task_status)
                # Output: finished
        """
        status_str = self.get_info_by_id(task_id)[ApiField.STATUS]  # @TODO: convert json to tuple
        return self.Status(status_str)

    def raise_for_status(self, status: Status) -> None:
        """
        Raise error if Application Task status is ERROR.

        :param status: Status object.
        :type status: Status
        :returns: None
        :rtype: None
        """
        if status is self.Status.ERROR:
            raise TaskFinishedWithError(f"Task finished with status {str(self.Status.ERROR)}")

    def wait(
        self,
        id: int,
        target_status: Status,
        wait_attempts: Optional[int] = None,
        wait_attempt_timeout_sec: Optional[int] = None,
    ):
        """
        Awaiting achievement by given Application Task of a given status.

        :param id: Task ID in Supervisely.
        :type id: int
        :param target_status: Status object(status of task we expect to destinate).
        :type target_status: Status
        :param wait_attempts: The number of attempts to determine the status of the task that we are waiting for.
        :type wait_attempts: int, optional
        :param wait_attempt_timeout_sec: Number of seconds for intervals between attempts(raise error if waiting time exceeded).
        :type wait_attempt_timeout_sec: int, optional
        :returns: True if the desired status is reached, False otherwise
        :rtype: bool
        """
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status in [
                target_status,
                self.Status.FINISHED,
                self.Status.DEPLOYED,
                self.Status.STOPPED,
            ]:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded(
            f"Waiting time exceeded: total waiting time {wait_attempts * effective_wait_timeout} seconds, i.e. {wait_attempts} attempts for {effective_wait_timeout} seconds each"
        )

    def get_context(self, id: int) -> Dict:
        """
        Get context information by Application Task ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :returns: Context information in dict format
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                task_id = 121230
                context = api.task.get_context(task_id)
                print(context)
                # Output: {
                #     "team": {
                #         "id": 16087,
                #         "name": "alexxx"
                #     },
                #     "workspace": {
                #         "id": 23821,
                #         "name": "my_super_workspace"
                #     }
                # }
        """
        response = self._api.post("GetTaskContext", {ApiField.ID: id})
        return response.json()

    def _convert_json_info(self, info: dict):
        """_convert_json_info"""
        return info

    def start(
        self,
        agent_id,
        app_id: Optional[int] = None,
        workspace_id: Optional[int] = None,
        description: Optional[str] = "application description",
        params: Dict[str, Any] = None,
        log_level: Optional[Literal["info", "debug", "warning", "error"]] = "info",
        users_ids: Optional[List[int]] = None,
        app_version: Optional[str] = "",
        is_branch: Optional[bool] = False,
        task_name: Optional[str] = "pythonSpawned",
        restart_policy: Optional[Literal["never", "on_error"]] = "never",
        proxy_keep_url: Optional[bool] = False,
        module_id: Optional[int] = None,
        redirect_requests: Optional[Dict[str, int]] = {},
        limit_by_workspace: bool = False,
        kubernetes_settings: Optional[Union[KubernetesSettings, Dict[str, Any]]] = None,
        multi_user_session: bool = False,
    ) -> Dict[str, Any]:
        """Starts the Application Task on the agent.

        :param agent_id: Agent ID. Can be obtained from TeamCluster page in UI.
        :type agent_id: int
        :param app_id: Deprecated. Use module_id instead.
        :type app_id: int, optional
        :param workspace_id: Workspace ID where the task will be created.
        :type workspace_id: int, optional
        :param description: Task description which will be shown in UI.
        :type description: str, optional
        :param params: Task parameters which will be passed to the application, check the
            code example below for more details.
        :type params: Dict[str, Any], optional
        :param log_level: Log level for the application.
        :type log_level: Literal["info", "debug", "warning", "error"], optional
        :param users_ids: List of user IDs for which will be created an instance of the application.
            For each user a separate task will be created.
        :type users_ids: List[int], optional
        :param app_version: Application version e.g. "v1.0.0" or branch name e.g. "dev".
        :type app_version: str, optional
        :param is_branch: If the application version is a branch name, set this parameter to True.
        :type is_branch: bool, optional
        :param task_name: Task name which will be shown in UI.
        :type task_name: str, optional
        :param restart_policy: when the task should be restarted: never or if error occurred.
        :type restart_policy: Literal["never", "on_error"], optional
        :param proxy_keep_url: For internal usage only.
        :type proxy_keep_url: bool, optional
        :param module_id: Module ID. Can be obtained from the apps page in UI.
        :type module_id: int, optional
        :param redirect_requests: For internal usage only in Develop and Debug mode.
        :type redirect_requests: Dict[str, int], optional
        :param limit_by_workspace: If set to True tasks will be only visible inside of the workspace
            with specified workspace_id.
        :type limit_by_workspace: bool, optional
        :param kubernetes_settings: Kubernetes settings for the application.
        :type kubernetes_settings: Union[:class:`~supervisely.api.task_api.KubernetesSettings`, Dict[str, Any]], optional
        :param multi_user_session: If True, the application session will be created as multi-user.
                                   In this case, multiple users will be able to connect to the same application session.
                                   All users will have separate application states.
                                   Available only for applications that support multi-user sessions.
        :type multi_user_session: bool, default is False
        :returns: Task information in JSON format.
        :rtype: Dict[str, Any]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                app_slug = "supervisely-ecosystem/export-to-supervisely-format"
                module_id = api.app.get_ecosystem_module_id(app_slug)
                module_info = api.app.get_ecosystem_module_info(module_id)

                project_id = 12345
                agent_id = 12345
                workspace_id = 12345

                params = module_info.get_arguments(images_project=project_id)

                session = api.app.start(
                    agent_id=agent_id,
                    module_id=module_id,
                    workspace_id=workspace_id,
                    task_name="Prepare download link",
                    params=params,
                    app_version="dninja",
                    is_branch=True,
                )
        """
        if app_id is not None and module_id is not None:
            raise ValueError("Only one of the arguments (app_id or module_id) have to be defined")
        if app_id is None and module_id is None:
            raise ValueError("One of the arguments (app_id or module_id) have to be defined")

        advanced_settings = {
            ApiField.LIMIT_BY_WORKSPACE: limit_by_workspace,
        }

        if kubernetes_settings is not None:
            if isinstance(kubernetes_settings, KubernetesSettings):
                kubernetes_settings = kubernetes_settings.to_dict()
            if not isinstance(kubernetes_settings, dict):
                raise TypeError(
                    f"kubernetes_settings must be a dict or an instance of KubernetesSettings, got {type(kubernetes_settings)}"
                )
            advanced_settings.update(kubernetes_settings)

        data = {
            ApiField.AGENT_ID: agent_id,
            # "nodeId": agent_id,
            ApiField.WORKSPACE_ID: workspace_id,
            ApiField.DESCRIPTION: description,
            ApiField.PARAMS: take_with_default(params, {"state": {}}),
            ApiField.LOG_LEVEL: log_level,
            ApiField.USERS_IDS: take_with_default(users_ids, []),
            ApiField.APP_VERSION: app_version,
            ApiField.IS_BRANCH: is_branch,
            ApiField.TASK_NAME: task_name,
            ApiField.RESTART_POLICY: restart_policy,
            ApiField.PROXY_KEEP_URL: proxy_keep_url,
            ApiField.ADVANCED_SETTINGS: advanced_settings,
        }
        if len(redirect_requests) > 0:
            data[ApiField.REDIRECT_REQUESTS] = redirect_requests

        if app_id is not None:
            data[ApiField.APP_ID] = app_id
        if module_id is not None:
            data[ApiField.MODULE_ID] = module_id
        if multi_user_session:
            # * Enables single multi-user session mode for all users in the users_ids list.
            # * Otherwise, if users_ids contains multiple IDs, separate single-user sessions will be created for each.
            # * If users_ids is empty, a session is created only for the current user.
            data[ApiField.SINGLE_SESSION_MODE] = multi_user_session
        resp = self._api.post(method="tasks.run.app", data=data)
        task = resp.json()[0]
        if "id" not in task:
            task["id"] = task.get("taskId")
        return task

    def stop(self, id: int):
        """stop"""
        response = self._api.post("tasks.stop", {ApiField.ID: id})
        return self.Status(response.json()[ApiField.STATUS])

    def submit_logs(self, logs) -> None:
        """submit_logs"""
        response = self._api.post("tasks.logs.add", {ApiField.LOGS: logs})
        # return response.json()[ApiField.TASK_ID]

    def upload_files(
        self,
        task_id: int,
        abs_paths: List[str],
        names: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """upload_files"""
        if len(abs_paths) != len(names):
            raise RuntimeError("Inconsistency: len(abs_paths) != len(names)")

        hashes = []
        if len(abs_paths) == 0:
            return

        hash_to_items = defaultdict(list)
        hash_to_name = defaultdict(list)
        for idx, item in enumerate(zip(abs_paths, names)):
            path, name = item
            item_hash = get_file_hash(path)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(path)
            hash_to_name[item_hash].append(name)

        unique_hashes = set(hashes)
        remote_hashes = self._api.image.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        # @TODO: upload remote hashes
        if len(remote_hashes) != 0:
            files = []
            for hash in remote_hashes:
                for name in hash_to_name[hash]:
                    files.append({ApiField.NAME: name, ApiField.HASH: hash})
            for batch in batched(files):
                resp = self._api.post(
                    "tasks.files.bulk.add-by-hash",
                    {ApiField.TASK_ID: task_id, ApiField.FILES: batch},
                )
        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        for batch in batched(list(zip(abs_paths, names, hashes))):
            content_dict = OrderedDict()
            for idx, item in enumerate(batch):
                path, name, hash = item
                if hash in remote_hashes:
                    continue
                content_dict["{}".format(idx)] = json.dumps({"fullpath": name, "hash": hash})
                content_dict["{}-file".format(idx)] = (name, open(path, "rb"), "")

            if len(content_dict) > 0:
                encoder = MultipartEncoder(fields=content_dict)
                resp = self._api.post("tasks.files.bulk.upload", encoder)
                if progress_cb is not None:
                    progress_cb(len(content_dict))

    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2}
    # }
    # {
    #     obj: {new_val: 1}
    # }
    # // apped: true, recursive: false
    # {
    #     data: {my_val: 1}
    #     obj: {new_val: 1}
    # }(edited)
    # // append: false, recursive: false
    # {
    #     obj: {new_val: 1}
    # }(edited)
    #
    # 16: 32
    # // append: true, recursive: true
    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2, new_val: 1}
    # }

    def set_fields(self, task_id: int, fields: List) -> Dict:
        """set_fields"""
        for idx, obj in enumerate(fields):
            for key in [ApiField.FIELD, ApiField.PAYLOAD]:
                if key not in obj:
                    raise KeyError("Object #{} does not have field {!r}".format(idx, key))
        data = {ApiField.TASK_ID: task_id, ApiField.FIELDS: fields}
        resp = self._api.post("tasks.data.set", data)
        return resp.json()

    def set_fields_from_dict(self, task_id: int, d: Dict) -> Dict:
        """set_fields_from_dict"""
        fields = []
        for k, v in d.items():
            fields.append({ApiField.FIELD: k, ApiField.PAYLOAD: v})
        return self.set_fields(task_id, fields)

    def set_field(
        self,
        task_id: int,
        field: Dict,
        payload: Dict,
        append: Optional[bool] = False,
        recursive: Optional[bool] = False,
    ) -> Dict:
        """set_field"""
        fields = [
            {
                ApiField.FIELD: field,
                ApiField.PAYLOAD: payload,
                ApiField.APPEND: append,
                ApiField.RECURSIVE: recursive,
            }
        ]
        return self.set_fields(task_id, fields)

    def get_fields(self, task_id, fields: List):
        """get_fields"""
        data = {ApiField.TASK_ID: task_id, ApiField.FIELDS: fields}
        resp = self._api.post("tasks.data.get", data)
        return resp.json()["result"]

    def get_field(self, task_id: int, field: str):
        """get_field"""
        result = self.get_fields(task_id, [field])
        return result[field]

    def _set_output(self):
        """_set_output"""
        pass

    def set_output_project(
        self,
        task_id: int,
        project_id: int,
        project_name: Optional[str] = None,
        project_preview: Optional[str] = None,
    ) -> Dict:
        """set_output_project"""
        if "import" in app_categories():
            self._api.project.add_import_history(project_id, task_id)
        if project_name is None:
            project = self._api.project.get_info_by_id(project_id, raise_error=True)
            project_name = project.name
            project_preview = project.image_preview_url

        output = {ApiField.PROJECT: {ApiField.ID: project_id, ApiField.TITLE: project_name}}
        if project_preview is not None:
            output[ApiField.PROJECT][ApiField.PREVIEW] = project_preview
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def set_output_report(
        self,
        task_id: int,
        file_id: int,
        file_name: str,
        description: Optional[str] = "Report",
    ) -> Dict:
        """set_output_report"""
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            description=description,
            icon="zmdi zmdi-receipt",
        )

    def _set_custom_output(
        self,
        task_id,
        file_id,
        file_name,
        file_url=None,
        description="File",
        icon="zmdi zmdi-file-text",
        color="#33c94c",
        background_color="#d9f7e4",
        download=False,
    ):
        """_set_custom_output"""
        if file_url is None:
            file_url = self._api.file.get_url(file_id)

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": icon,
                    "color": color,
                    "backgroundColor": background_color,
                },
                "title": file_name,
                "titleUrl": file_url,
                "download": download,
                "description": description,
            }
        }
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def set_output_archive(
        self, task_id: int, file_id: int, file_name: str, file_url: Optional[str] = None
    ) -> Dict:
        """set_output_archive"""
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            file_url=file_url,
            description="Download archive",
            icon="zmdi zmdi-archive",
            download=True,
        )

    def set_output_file_download(
        self,
        task_id: int,
        file_id: int,
        file_name: str,
        file_url: Optional[str] = None,
        download: Optional[bool] = True,
    ) -> Dict:
        """set_output_file_download"""
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            file_url=file_url,
            description="Download file",
            icon="zmdi zmdi-file",
            download=download,
        )

    def send_request(
        self,
        task_id: int,
        method: str,
        data: Dict,
        context: Optional[Dict] = {},
        skip_response: bool = False,
        timeout: Optional[int] = 60,
        outside_request: bool = True,
        retries: int = 10,
        raise_error: bool = False,
    ):
        """send_request"""
        if type(data) is not dict:
            raise TypeError("data argument has to be a dict")
        context["outside_request"] = outside_request
        resp = self._api.post(
            "tasks.request.direct",
            {
                ApiField.TASK_ID: task_id,
                ApiField.COMMAND: method,
                ApiField.CONTEXT: context,
                ApiField.STATE: data,
                "skipResponse": skip_response,
                "timeout": timeout,
            },
            retries=retries,
            raise_error=raise_error,
        )
        return resp.json()

    def set_output_directory(self, task_id, file_id, directory_path):
        """set_output_directory"""
        return self._set_custom_output(
            task_id,
            file_id,
            directory_path,
            description="Directory",
            icon="zmdi zmdi-folder",
        )

    def update_meta(
        self,
        id: int,
        data: dict,
        agent_storage_folder: str = None,
        relative_app_dir: str = None,
    ):
        """
        Update given task metadata
        :param id: int - task id
        :param data: dict - meta data to update
        """
        if type(data) == dict:
            data.update({"id": id})
            if agent_storage_folder is None and relative_app_dir is not None:
                raise ValueError(
                    "Both arguments (agent_storage_folder and relative_app_dir) has to be defined or None"
                )
            if agent_storage_folder is not None and relative_app_dir is None:
                raise ValueError(
                    "Both arguments (agent_storage_folder and relative_app_dir) has to be defined or None"
                )
            if agent_storage_folder is not None and relative_app_dir is not None:
                data["agentStorageFolder"] = {
                    "hostDir": agent_storage_folder,
                    "folder": relative_app_dir,
                }

        self._api.post("tasks.meta.update", data)

    def _update_app_content(self, task_id: int, data_patch: List[Dict] = None, state: Dict = None):
        payload = {}
        if data_patch is not None and len(data_patch) > 0:
            payload[ApiField.DATA] = data_patch
        if state is not None and len(state) > 0:
            payload[ApiField.STATE] = state

        resp = self._api.post(
            "tasks.app-v2.data.set",
            {ApiField.TASK_ID: task_id, ApiField.PAYLOAD: payload},
        )
        return resp.json()

    def set_output_error(
        self,
        task_id: int,
        title: str,
        description: Optional[str] = None,
        show_logs: Optional[bool] = True,
    ) -> Dict:
        """
        Set custom error message to the task output.

        :param task_id: Application task ID.
        :type task_id: int
        :param title: Error message to be displayed in the task output.
        :type title: str
        :param description: Description to be displayed in the task output.
        :type description: Optional[str]
        :param show_logs: If True, the link to the task logs will be displayed in the task output.
        :type show_logs: Optional[bool], default True
        :returns: Response JSON.
        :rtype: Dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                task_id = 12345
                title = "Something went wrong"
                description = "Please check the task logs"
                show_logs = True
                api.task.set_output_error(task_id, title, description, show_logs)
        """

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": "zmdi zmdi-alert-octagon",
                    "color": "#ff83a6",
                    "backgroundColor": "#ffeae9",
                },
                "title": title,
                "showLogs": show_logs,
                "isError": True,
            }
        }

        if description is not None:
            output[ApiField.GENERAL]["description"] = description

        resp = self._api.post(
            "tasks.output.set",
            {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output},
        )
        return resp.json()

    def set_output_text(
        self,
        task_id: int,
        title: str,
        description: Optional[str] = None,
        show_logs: Optional[bool] = False,
        zmdi_icon: Optional[str] = "zmdi-comment-alt-text",
        icon_color: Optional[str] = "#33c94c",
        background_color: Optional[str] = "#d9f7e4",
    ) -> Dict:
        """
        Set custom text message to the task output.

        :param task_id: Application task ID.
        :type task_id: int
        :param title: Text message to be displayed in the task output.
        :type title: str
        :param description: Description to be displayed in the task output.
        :type description: Optional[str]
        :param show_logs: If True, the link to the task logs will be displayed in the task output.
        :type show_logs: Optional[bool], default False
        :param zmdi_icon: Icon class name from Material Design Icons set (ZMDI).
        :type zmdi_icon: Optional[str], default "zmdi-comment-alt-text"
        :param icon_color: Icon color in HEX format.
        :type icon_color: Optional[str], default "#33c94c" (nearest Duron Jolly Green)
        :param background_color: Background color in HEX format.
        :type background_color: Optional[str], default "#d9f7e4" (Cosmic Latte)
        :returns: Response JSON.
        :rtype: Dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                task_id = 12345
                title = "Task is finished"
                api.task.set_output_text(task_id, title)
        """

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": f"zmdi {zmdi_icon}",
                    "color": icon_color,
                    "backgroundColor": background_color,
                },
                "title": title,
                "showLogs": show_logs,
                "isError": False,
            }
        }

        if description is not None:
            output[ApiField.GENERAL]["description"] = description

        resp = self._api.post(
            "tasks.output.set",
            {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output},
        )
        return resp.json()

    def update_status(
        self,
        task_id: int,
        status: Status,
    ) -> None:
        """Sets the specified status for the task.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :param status: Task status to set.
        :type status: One of the values from :class:`Status`, e.g. Status.FINISHED, Status.ERROR, etc.
        :raises ValueError: If the status value is not allowed.
        """
        # If status was passed without converting to string, convert it.
        # E.g. Status.FINISHED -> "finished"
        status = str(status)
        if status not in self.Status.values():
            raise ValueError(
                f"Invalid status value: {status}. Allowed values: {self.Status.values()}"
            )
        self._api.post("tasks.status.update", {ApiField.ID: task_id, ApiField.STATUS: status})

    def set_output_experiment(self, task_id: int, experiment_info: dict) -> Dict:
        """
        Sets output for the task with experiment info.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :param experiment_info: Experiment info from :class:`~supervisely.nn.training.train_app.TrainApp`. See class :class:`~supervisely.nn.experiments.ExperimentInfo`.
        :type experiment_info: dict
        :returns: Server response JSON.
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                experiment_info = {
                    "experiment_name": "247 Lemons RT-DETRv2-M",
                    "framework_name": "RT-DETRv2",
                    "model_name": "RT-DETRv2-M",
                    "task_type": "object detection",
                    "project_id": 76,
                    "project_version": {"id": 222, "version": 4},
                    "task_id": 247,
                    "model_files": {"config": "model_config.yml"},
                    "checkpoints": [
                        "checkpoints/best.pth",
                        "checkpoints/checkpoint0025.pth",
                        "checkpoints/checkpoint0050.pth",
                        "checkpoints/last.pth",
                    ],
                    "best_checkpoint": "best.pth",
                    "export": {"ONNXRuntime": "export/best.onnx"},
                    "app_state": "app_state.json",
                    "model_meta": "model_meta.json",
                    "train_val_split": "train_val_split.json",
                    "train_size": 4,
                    "val_size": 2,
                    "train_collection_id": 530,
                    "val_collection_id": 531,
                    "hyperparameters": "hyperparameters.yaml",
                    "artifacts_dir": "/experiments/76_Lemons/247_RT-DETRv2/",
                    "datetime": "2025-01-22 18:13:43",
                    "evaluation_report_id": 12961,
                    "evaluation_report_link": "https://app.supervisely.com/model-benchmark?id=12961",
                    "evaluation_metrics": {"mAP": 0.994, "AP50": 1.0, "AP75": 1.0},  # and other metrics
                    "primary_metric": "mAP",
                    "logs": {"type": "tensorboard", "link": "/experiments/76_Lemons/247_RT-DETRv2/logs/"},
                }
        """
        output = {
            ApiField.EXPERIMENT: {ApiField.DATA: {**experiment_info}},
        }
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def is_running(self, task_id: int) -> bool:
        """
        Check if the task is running.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :returns: True if the task is running, False otherwise.
        :rtype: bool
        """
        try:
            self.send_request(task_id, "is_running", {}, retries=1, raise_error=True)
        except requests.exceptions.HTTPError as e:
            return False
        return True

    def is_ready(self, task_id: int) -> bool:
        """
        Check if the task is ready.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :returns: True if the task is ready, False otherwise.
        :rtype: bool
        """
        try:
            return (
                self.send_request(task_id, "is_ready", {}, retries=1, raise_error=True)["status"]
                == "ready"
            )
        except requests.exceptions.HTTPError as e:
            return False



# coding: utf-8

# docs
from __future__ import annotations

from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from supervisely.sly_logger import logger
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    ID,
    LABELER_LOGIN,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.geometry.point import Point
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.color import _validate_color, hex2rgb, rgb2hex
from supervisely.io.json import JsonSerializable

EDGES = "edges"
NODES = "nodes"

DISABLED = "disabled"
LOC = "loc"

DST = "dst"
SRC = "src"
COLOR = "color"


class Node(JsonSerializable):
    """Single vertex in a keypoint graph: a labeled point (row, col) used as a building block for :class:`~supervisely.geometry.graph.GraphNodes` (e.g. pose keypoints, skeleton joints)."""

    def __init__(
        self,
        location: Optional[PointLocation] = None,
        disabled: Optional[bool] = False,
        label: Optional[str] = None,
        row: Optional[int] = None,
        col: Optional[int] = None,
    ):
        """
        Node for a single :class:`~supervisely.geometry.graph.GraphNodes`.

        :param location: Location of Node.
        :type location: :class:`~supervisely.geometry.point_location.PointLocation`
        :param disabled: Determines whether to display the Node when drawing or not.
        :type disabled: bool, optional
        :param label: str
        :param row: int
        :param col: int

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.graph import Node

                vertex = Node(sly.PointLocation(5, 5))
        """
        if None not in (location, row, col) or all(item is None for item in (location, row, col)):
            raise ValueError("Either location or row and col must be specified")
        self._location = location
        self._disabled = disabled
        self._label = label
        if None not in [row, col]:
            self._location = PointLocation(row, col)

    @property
    def location(self) -> PointLocation:
        """
        Location of Node.

        :returns: Location of Node.
        :rtype: :class:`~supervisely.geometry.point_location.PointLocation`
        """
        return self._location

    @property
    def disabled(self) -> bool:
        """
        Display the Node when drawing or not.

        :returns: Boolean
        :rtype: bool
        """
        return self._disabled

    @classmethod
    def from_json(cls, data: Dict) -> Node:
        """
        Convert a json dict to Node. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Node in json format as a dict.
        :type data: dict
        :returns: Node from json.
        :rtype: :class:`~supervisely.geometry.graph.Node`

        :Usage Example:

            .. code-block:: python

                vertex_json = {"loc": [5, 5]}
                vertex = Node.from_json(vertex_json)
        """
        # TODO validations
        loc = data[LOC]
        return cls(
            location=PointLocation(row=loc[1], col=loc[0]),
            disabled=data.get(DISABLED, False),
        )

    def to_json(self) -> Dict:
        """
        Convert the Node to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Node in json format as a dict.
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.graph import Node

                vertex = Node(sly.PointLocation(5, 5))
                vertex_json = vertex.to_json()
                print(vertex_json)
                # Output: {
                #    "loc": [5, 5]
                # }
        """
        result = {LOC: [self._location.col, self._location.row]}
        if self.disabled:
            result[DISABLED] = True
        return result

    def transform_location(self, transform_fn):
        """
        :param transform_fn: function to convert location
        :returns: Node after transformation.
        :rtype: :class:`~supervisely.geometry.graph.Node`
        """
        return Node(transform_fn(self._location), disabled=self.disabled)


def _maybe_transform_colors(elements, process_fn):
    """
    Function _maybe_transform_colors convert some list of parameters using the given function
    :param elements: List of elements.
    :param process_fn: function to convert
    :type process_fn: function
    :returns: List of elements after transformation.
    :rtype: List[Any]
    """
    for elem in elements:
        if COLOR in elem:
            elem[COLOR] = process_fn(elem[COLOR])


class GraphNodes(Geometry):
    """Geometry representing a graph of labeled keypoints: a collection of :class:`~supervisely.geometry.graph.Node` objects for pose estimation, skeletons, or structured point sets (e.g. nose, left_eye, right_shoulder)."""

    items_json_field = NODES

    @staticmethod
    def geometry_name():
        return "graph"

    def __init__(
        self,
        nodes: Union[Dict[str, Dict], List],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        """
        GraphNodes geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.graph.GraphNodes` class object is immutable.

        :param nodes: Dict or List containing nodes of graph
        :type nodes: dict
        :param sly_id: GraphNodes ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which GraphNodes belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created :class:`~supervisely.geometry.graph.GraphNodes`.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when GraphNodes was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when GraphNodes was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.graph import Node, GraphNodes

                vertex_1 = Node(sly.PointLocation(5, 5))
                vertex_2 = Node(sly.PointLocation(100, 100))
                vertex_3 = Node(sly.PointLocation(200, 250))
                nodes = {0: vertex_1, 1: vertex_2, 2: vertex_3}
                figure = GraphNodes(nodes)
        """

        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        if len(nodes) == 0:
            raise ValueError("Empty list of nodes is not allowed for GraphNodes")
        self._nodes = nodes
        if isinstance(nodes, (list, tuple)):
            self._nodes = {}
            for i, node in enumerate(nodes):
                if node._label is not None:
                    self._nodes[node._label] = Node(node._location, node._disabled)
                else:
                    self._nodes[str(i)] = Node(node._location, node._disabled)

    @property
    def nodes(self) -> Dict[str, Dict]:
        """
        Copy of GraphNodes nodes.

        :returns: GraphNodes nodes
        :rtype: dict
        """
        return self._nodes.copy()

    @classmethod
    def from_json(cls, data: Dict[str, Dict]) -> GraphNodes:
        """
        Convert a json dict to GraphNodes. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: GraphNodes in json format as a dict.
        :type data: dict
        :returns: GraphNodes from json.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                figure_json = {
                    "nodes": {
                        "0": {"loc": [5, 5]},
                        "1": {"loc": [100, 100]},
                        "2": {"loc": [250, 200]}
                    }
                }
                from supervisely.geometry.graph import GraphNodes
                figure = GraphNodes.from_json(figure_json)
        """
        nodes = {
            node_id: Node.from_json(node_json)
            for node_id, node_json in data[cls.items_json_field].items()
        }
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(
            nodes=nodes,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def to_json(self) -> Dict[str, Dict]:
        """
        Convert the GraphNodes to list. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Json format as a dict
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.graph import Node, GraphNodes

                vertex_1 = Node(sly.PointLocation(5, 5))
                vertex_2 = Node(sly.PointLocation(100, 100))
                vertex_3 = Node(sly.PointLocation(200, 250))
                nodes = {0: vertex_1, 1: vertex_2, 2: vertex_3}
                figure = GraphNodes(nodes)

                figure_json = figure.to_json()
                print(figure_json)
                # Output: {
                #    "nodes": {
                #        "0": {
                #            "loc": [5, 5]
                #        },
                #        "1": {
                #            "loc": [100, 100]
                #        },
                #        "2": {
                #            "loc": [250, 200]
                #        }
                #    }
                # }
        """
        res = {
            self.items_json_field: {
                node_id: node.to_json() for node_id, node in self._nodes.items()
            }
        }
        self._add_creation_info(res)
        return res

    def crop(self, rect: Rectangle) -> List[GraphNodes]:
        """
        Crops current GraphNodes.

        :param rect: Rectangle for crop.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of GraphNodes after crop.
        :rtype: List[:class:`~supervisely.geometry.graph.GraphNodes`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                crop_figures = figure.crop(sly.Rectangle(0, 0, 300, 350))
        """
        is_all_nodes_inside = all(
            rect.contains_point_location(node.location) for node in self._nodes.values()
        )
        return [self] if is_all_nodes_inside else []

    def relative_crop(self, rect: Rectangle) -> List[GraphNodes]:
        """
        Crops current GraphNodes with given rectangle and shifts it on value of rectangle left top angle.

        :param rect: Rectangle for crop.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of GraphNodes after relative crop.
        :rtype: List[:class:`~supervisely.geometry.graph.GraphNodes`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                rel_crop_figures = figure.relative_crop(sly.Rectangle(0, 0, 300, 350))
        """
        return [geom.translate(drow=-rect.top, dcol=-rect.left) for geom in self.crop(rect)]

    def transform(self, transform_fn) -> GraphNodes:
        """
        :param transform_fn: Function to convert GraphNodes.
        :type transform_fn: function
        :returns: GraphNodes after transformation.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`
        """
        return self.__class__(
            nodes={node_id: transform_fn(node) for node_id, node in self._nodes.items()}
        )

    def transform_locations(self, transform_fn) -> GraphNodes:
        """
        :param transform_fn: Function to convert GraphNodes location.
        :type transform_fn: function
        :returns: GraphNodes after transformation.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`
        """
        return self.transform(lambda kp: kp.transform_location(transform_fn))

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> GraphNodes:
        """
        Resizes current GraphNodes.

        :param in_size: Input image size (height, width) to which belongs :class:`~supervisely.geometry.graph.GraphNodes`.
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) to which belongs :class:`~supervisely.geometry.graph.GraphNodes`.
        :type out_size: Tuple[int, int]
        :returns: GraphNodes after resize.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of Rectangle to a new variable
                in_height, in_width = 300, 400
                out_height, out_width = 600, 800
                resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        return self.transform_locations(lambda p: p.resize(in_size, out_size))

    def scale(self, factor: float) -> GraphNodes:
        """
        Scales current GraphNodes.

        :param factor: Scale parameter.
        :type factor: float
        :returns: :class:`~supervisely.geometry.graph.GraphNodes` object
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of Rectangle to a new variable
                scale_figure = figure.scale(0.75)
        """
        return self.transform_locations(lambda p: p.scale(factor))

    def translate(self, drow: int, dcol: int) -> GraphNodes:
        """
        Translates current GraphNodes.

        :param drow: Horizontal shift.
        :type drow: int
        :param dcol: Vertical shift.
        :type dcol: int
        :returns: GraphNodes after translate.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of Rectangle to a new variable
                translate_figure = figure.translate(150, 250)
        """
        return self.transform_locations(lambda p: p.translate(drow, dcol))

    def rotate(self, rotator: ImageRotator) -> GraphNodes:
        """
        Rotates current GraphNodes.

        :param rotator: Class for image rotation.
        :type rotator: :class:`~supervisely.geometry.image_rotator.ImageRotator`
        :returns: GraphNodes after rotation.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                from supervisely.geometry.image_rotator import ImageRotator

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of Rectangle to a new variable
                height, width = 300, 400
                rotator = ImageRotator((height, width), 25)
                rotate_figure = figure.rotate(rotator)
        """
        return self.transform_locations(lambda p: p.rotate(rotator))

    def fliplr(self, img_size: Tuple[int, int]) -> GraphNodes:
        """
        Flips current GraphNodes in horizontal.

        :param img_size: Input image size (height, width) to which belongs :class:`~supervisely.geometry.graph.GraphNodes`.
        :type img_size: Tuple[int, int]
        :returns: GraphNodes after flip in horizontal.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of Rectangle to a new variable
                height, width = 300, 400
                fliplr_figure = figure.fliplr((height, width))
        """
        return self.transform_locations(lambda p: p.fliplr(img_size))

    def flipud(self, img_size: Tuple[int, int]) -> GraphNodes:
        """
        Flips current GraphNodes in vertical.

        :param img_size: Input image size (height, width).
        :type img_size: Tuple[int, int]
        :returns: GraphNodes after flip in vertical.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of Rectangle to a new variable
                height, width = 300, 400
                flipud_figure = figure.flipud((height, width))
        """
        return self.transform_locations(lambda p: p.flipud(img_size))

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """
        Draws the graph contour on a given bitmap canvas
        :param bitmap: numpy array
        :param color: tuple or list of integers
        :param thickness: int
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        """
        self.draw_contour(bitmap, color, thickness, config=config)

    @staticmethod
    def _get_nested_or_default(dict, keys_path, default=None):
        """
        _get_nested_or_default
        """
        result = dict
        for key in keys_path:
            if result is not None:
                result = result.get(key, None)
        return result if result is not None else default

    def _draw_contour_impl(self, bitmap, color=None, thickness=1, config=None):
        """
        _draw_contour_impl
        """
        if config is not None:
            # If a config with edges and colors is passed, make sure it is
            # consistent with the our set of points.
            self.validate(self.geometry_name(), config)

        # Draw edges first so that nodeas are then drawn on top.
        for edge in self._get_nested_or_default(config, [EDGES], []):
            src = self._nodes.get(edge[SRC], None)
            dst = self._nodes.get(edge[DST], None)
            if (
                (src is not None)
                and (not src.disabled)
                and (dst is not None)
                and (not dst.disabled)
            ):
                edge_color = edge.get(COLOR, color)
                cv2.line(
                    bitmap,
                    (src.location.col, src.location.row),
                    (dst.location.col, dst.location.row),
                    tuple(edge_color),
                    thickness,
                )

        nodes_config = self._get_nested_or_default(config, [self.items_json_field])
        for node_id, node in self._nodes.items():
            if not node.disabled:
                effective_color = self._get_nested_or_default(nodes_config, [node_id, COLOR], color)
                Point.from_point_location(node.location).draw(
                    bitmap=bitmap,
                    color=effective_color,
                    thickness=thickness,
                    config=None,
                )

    @property
    def area(self) -> float:
        """
        GraphNodes area.

        :returns: Area of current GraphNodes, always 0.0
        :rtype: float

        :Usage Example:

            .. code-block:: python

                print(figure.area)
                # Output: 0.0
        """
        return 0.0

    def to_bbox(self) -> Rectangle:
        """
        Create Rectangle object from current GraphNodes.

        :returns: Rectangle from GraphNodes.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                rectangle = figure.to_bbox()
        """
        if self._nodes is None or len(self._nodes) == 0:
            logger.warning(
                f"Cannot create a bounding box from {self.name()} with empty nodes. Geometry ID: {self.sly_id} "
            )
        return Rectangle.from_geometries_list(
            [Point.from_point_location(node.location) for node in self._nodes.values()]
        )

    def clone(self) -> GraphNodes:
        """
        Makes a copy of the GraphNodes.

        :returns: Copy of GraphNodes.
        :rtype: :class:`~supervisely.geometry.graph.GraphNodes`

        :Usage Example:

            .. code-block:: python

                # Remember that GraphNodes class object is immutable, and we need to assign new instance of PointLocation to a new variable
                new_figure = figure.clone()
        """
        return self

    def validate(self, name: str, settings: Dict) -> None:
        """
        Checks the graph for correctness and compliance with the template

        :param name: Name of the geometry.
        :type name: str
        :param settings: Settings of the geometry.
        :type settings: dict
        :raises ValueError: if graph contains nodes not declared in the template
        """
        super().validate(name, settings)
        # TODO template self-consistency checks.

        nodes_not_in_template = set(self._nodes.keys()) - set(
            settings[self.items_json_field].keys()
        )
        if len(nodes_not_in_template) > 0:
            raise ValueError(
                "Graph contains nodes not declared in the template: {!r}.".format(
                    nodes_not_in_template
                )
            )

    @staticmethod
    def _transform_config_colors(config, transform_fn):
        """
        Transform colors of edges and nodes in graph template
        :param config: dictionary(graph template)
        :param transform_fn: function to convert
        :type transform_fn: function
        :returns: dictionary(graph template) after transformation
        :rtype: dict
        """
        if config is None:
            return None

        result = deepcopy(config)
        _maybe_transform_colors(result.get(EDGES, []), transform_fn)
        _maybe_transform_colors(result[NODES].values(), transform_fn)
        return result

    @staticmethod
    def config_from_json(config: Dict) -> Dict:
        """
        Convert graph template from json format

        :param config: dictionary(graph template) in json format
        :type config: dict
        :returns: dictionary(graph template) in json format
        :rtype: dict
        """

        try:
            return GraphNodes._transform_config_colors(config, hex2rgb)
        except Exception as e:
            raise RuntimeError(
                f"Failed to parse graph template from JSON format. "
                "Check out an example of a graph template in JSON format at: "
                "https://developer.supervisely.com/getting-started/python-sdk-tutorials/images/keypoints#click-to-see-the-example-of-template-in-json-format"
            )

    @staticmethod
    def config_to_json(config: Dict) -> Dict:
        """
        Convert graph template to json format

        :param config: dictionary(graph template)
        :type config: dict
        :returns: dictionary(graph template) in json format
        :rtype: dict
        """
        return GraphNodes._transform_config_colors(config, rgb2hex)

    @classmethod
    def allowed_transforms(cls):
        """
        Returns the allowed transforms for the GraphNodes.
        """
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.rectangle import Rectangle

        return [AnyGeometry, Rectangle]


class KeypointsTemplate(GraphNodes, Geometry):
    """Graph geometry for defining keypoint templates (nodes, edges, labels) used in pose estimation."""

    def __init__(self):
        self._config = {self.items_json_field: {}, EDGES: []}
        self._point_names = []

    def add_point(self, label: str, row: int, col: int, color: list = [0, 0, 255]):
        _validate_color(color)
        if label in self._config[self.items_json_field]:
            raise KeyError(f"Label {label} already exists in the graph")
        self._point_names.append(label)
        self._config[self.items_json_field][label] = {
            "label": label,
            "loc": [row, col],
            "color": color,
        }

    def add_edge(self, src: str, dst: str, color: list = [0, 255, 0]):
        _validate_color(color)
        for elem in (src, dst):
            if elem not in self._config[self.items_json_field]:
                raise ValueError(f"There is no such node in the graph: {elem}")
        self._config[EDGES].append({"src": src, "dst": dst, "color": color})

    def get_nodes(self):
        self._nodes = {}
        for node in self._config[self.items_json_field]:
            loc = self._config[self.items_json_field][node]["loc"]
            self._nodes[node] = Node(PointLocation(loc[1], loc[0]))

    def draw(self, image: np.ndarray, thickness=7):
        self.get_nodes()
        self._draw_bool_compatible(
            self._draw_impl,
            bitmap=image,
            color=[0, 255, 0],
            thickness=thickness,
            config=self._config,
        )

    def to_json(self):
        return self.config_to_json(self._config)

    @property
    def config(self):
        return self._config

    @property
    def point_names(self):
        """
        Return point names in order in which they were added
        """
        return self._point_names



from typing import Callable, List, Optional, Union

from tqdm import tqdm

from supervisely import get_project_class
from supervisely.api.api import Api
from supervisely.project import read_project


def upload(
    src_dir: str,
    api: Api,
    workspace_id: int,
    project_name: Optional[str] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    **kwargs,
) -> None:
    """
    Uploads project of any modality from the local directory. See methods `sly.upload_project`,
    `sly.upload_video_project`, `sly.upload_volume_project`, `sly.upload_pointcloud_project`,
    `sly.upload_pointcloud_episode_project` to examine full list of possible arguments.

    :param src_dir: Source path to local directory.
    :type src_dir: str
    :param api: Supervisely API address and token.
    :type api: :class:`~supervisely.api.api.Api`
    :param workspace_id: Destination workspace ID.
    :type workspace_id: int
    :param project_name: Custom project name. By default, it's a directory name.
    :type project_name: str, optional
    :param log_progress: Show uploading logs in the output.
    :type log_progress: bool
    :param progress_cb: Function for tracking upload progress.
    :type progress_cb: tqdm or callable, optional

    :returns: None.
    :rtype: NoneType

    :Usage Example:

        .. code-block:: python

            import os
            from tqdm import tqdm
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))

            api = sly.Api.from_env()

            src_dir = '/your/local/source/dir'

            # Upload image project
            project_fs = sly.read_project(src_dir)

            pbar = tqdm(desc="Uploading image project", total=project_fs.total_items)
            sly.upload(src_dir, api, workspace_id, project_name, progress_cb=pbar)

            # Upload video project
            sly.upload(
                src_dir,
                api,
                workspace_id,
                project_name="Some Video Project",
                log_progress=True,
                include_custom_data=True
            )

            # Upload volume project
            sly.upload(src_dir, api, workspace_id, project_name="Some Volume Project", log_progress=True)

            # Upload pointcloud project
            project_fs = read_project(directory)

            pbar = tqdm(desc="Uploading pointcloud project", total=project_fs.total_items)
            sly.upload(
                src_dir,
                api,
                workspace_id,
                project_name="Some Pointcloud Project",
                progress_cb=pbar,
            )

            # Upload pointcloud episodes project
            project_fs = read_project(src_dir)

            with tqdm(desc="Upload pointcloud episodes project", total=project_fs.total_items) as pbar:
                sly.upload(
                    src_dir,
                    api,
                    workspace_id,
                    project_name="Some Pointcloud Episodes Project",
                    progress_cb=pbar,
                )
    """

    project_fs = read_project(src_dir)

    if progress_cb is not None:
        log_progress = False

    project_class = get_project_class(project_fs.meta.project_type)
    project_class.upload(
        src_dir,
        api=api,
        workspace_id=workspace_id,
        project_name=project_name,
        log_progress=log_progress,
        progress_cb=progress_cb,
        **kwargs,
    )
